//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|
//      \    \  |    |   ________________________________________________________________
//       \    \ |    |  |  ______________________________________________________________|
//        \    \|    |  |  |         __          __     __     __     ______      _______
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  \
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file CalcConc27.cu
//! \ingroup GPU
//! \author Martin Schoenherr
//=======================================================================================
/* Device code */
#include "LBM/LB.h"
#include "LBM/D3Q27.h"
#include <lbm/constants/NumericConstants.h>

using namespace vf::lbm::constant;

////////////////////////////////////////////////////////////////////////////////
extern "C" __global__ void CalcConc27(
	real* concentration,
	uint* typeOfGridNode,
	uint* neighborX,
	uint* neighborY,
	uint* neighborZ,
	uint size_Mat,
	real* distributionsAD,
	bool isEvenTimestep)
{
   //////////////////////////////////////////////////////////////////////////
   //! The velocity boundary condition is executed in the following steps
   //!
   ////////////////////////////////////////////////////////////////////////////////
   //! - Get node index coordinates from thredIdx, blockIdx, blockDim and gridDim.
   //!
   const unsigned  x = threadIdx.x;  // global x-index
   const unsigned  y = blockIdx.x;   // global y-index
   const unsigned  z = blockIdx.y;   // global z-index

   const unsigned nx = blockDim.x;
   const unsigned ny = gridDim.x;

   const unsigned k = nx*(ny*z + y) + x;
   //////////////////////////////////////////////////////////////////////////

   //////////////////////////////////////////////////////////////////////////
   // run for all indices in size_Mat and fluid nodes
   if ((k < size_Mat) && (typeOfGridNode[k] == GEO_FLUID))
   {
      //////////////////////////////////////////////////////////////////////////
      //! - Read distributions: style of reading and writing the distributions from/to stored arrays dependent on timestep is based on the esoteric twist algorithm \ref
      //! <a href="https://doi.org/10.3390/computation5020019"><b>[ M. Geier et al. (2017), DOI:10.3390/computation5020019 ]</b></a>
      //!
      Distributions27 distAD;
      if (isEvenTimestep)
      {
         distAD.f[dirE   ] = &distributionsAD[dirE   *size_Mat];
         distAD.f[dirW   ] = &distributionsAD[dirW   *size_Mat];
         distAD.f[dirN   ] = &distributionsAD[dirN   *size_Mat];
         distAD.f[dirS   ] = &distributionsAD[dirS   *size_Mat];
         distAD.f[dirT   ] = &distributionsAD[dirT   *size_Mat];
         distAD.f[dirB   ] = &distributionsAD[dirB   *size_Mat];
         distAD.f[dirNE  ] = &distributionsAD[dirNE  *size_Mat];
         distAD.f[dirSW  ] = &distributionsAD[dirSW  *size_Mat];
         distAD.f[dirSE  ] = &distributionsAD[dirSE  *size_Mat];
         distAD.f[dirNW  ] = &distributionsAD[dirNW  *size_Mat];
         distAD.f[dirTE  ] = &distributionsAD[dirTE  *size_Mat];
         distAD.f[dirBW  ] = &distributionsAD[dirBW  *size_Mat];
         distAD.f[dirBE  ] = &distributionsAD[dirBE  *size_Mat];
         distAD.f[dirTW  ] = &distributionsAD[dirTW  *size_Mat];
         distAD.f[dirTN  ] = &distributionsAD[dirTN  *size_Mat];
         distAD.f[dirBS  ] = &distributionsAD[dirBS  *size_Mat];
         distAD.f[dirBN  ] = &distributionsAD[dirBN  *size_Mat];
         distAD.f[dirTS  ] = &distributionsAD[dirTS  *size_Mat];
         distAD.f[dirREST] = &distributionsAD[dirREST*size_Mat];
         distAD.f[dirTNE ] = &distributionsAD[dirTNE *size_Mat];
         distAD.f[dirTSW ] = &distributionsAD[dirTSW *size_Mat];
         distAD.f[dirTSE ] = &distributionsAD[dirTSE *size_Mat];
         distAD.f[dirTNW ] = &distributionsAD[dirTNW *size_Mat];
         distAD.f[dirBNE ] = &distributionsAD[dirBNE *size_Mat];
         distAD.f[dirBSW ] = &distributionsAD[dirBSW *size_Mat];
         distAD.f[dirBSE ] = &distributionsAD[dirBSE *size_Mat];
         distAD.f[dirBNW ] = &distributionsAD[dirBNW *size_Mat];
      }
      else
      {
         distAD.f[dirW   ] = &distributionsAD[dirE   *size_Mat];
         distAD.f[dirE   ] = &distributionsAD[dirW   *size_Mat];
         distAD.f[dirS   ] = &distributionsAD[dirN   *size_Mat];
         distAD.f[dirN   ] = &distributionsAD[dirS   *size_Mat];
         distAD.f[dirB   ] = &distributionsAD[dirT   *size_Mat];
         distAD.f[dirT   ] = &distributionsAD[dirB   *size_Mat];
         distAD.f[dirSW  ] = &distributionsAD[dirNE  *size_Mat];
         distAD.f[dirNE  ] = &distributionsAD[dirSW  *size_Mat];
         distAD.f[dirNW  ] = &distributionsAD[dirSE  *size_Mat];
         distAD.f[dirSE  ] = &distributionsAD[dirNW  *size_Mat];
         distAD.f[dirBW  ] = &distributionsAD[dirTE  *size_Mat];
         distAD.f[dirTE  ] = &distributionsAD[dirBW  *size_Mat];
         distAD.f[dirTW  ] = &distributionsAD[dirBE  *size_Mat];
         distAD.f[dirBE  ] = &distributionsAD[dirTW  *size_Mat];
         distAD.f[dirBS  ] = &distributionsAD[dirTN  *size_Mat];
         distAD.f[dirTN  ] = &distributionsAD[dirBS  *size_Mat];
         distAD.f[dirTS  ] = &distributionsAD[dirBN  *size_Mat];
         distAD.f[dirBN  ] = &distributionsAD[dirTS  *size_Mat];
         distAD.f[dirREST] = &distributionsAD[dirREST*size_Mat];
         distAD.f[dirTNE ] = &distributionsAD[dirBSW *size_Mat];
         distAD.f[dirTSW ] = &distributionsAD[dirBNE *size_Mat];
         distAD.f[dirTSE ] = &distributionsAD[dirBNW *size_Mat];
         distAD.f[dirTNW ] = &distributionsAD[dirBSE *size_Mat];
         distAD.f[dirBNE ] = &distributionsAD[dirTSW *size_Mat];
         distAD.f[dirBSW ] = &distributionsAD[dirTNE *size_Mat];
         distAD.f[dirBSE ] = &distributionsAD[dirTNW *size_Mat];
         distAD.f[dirBNW ] = &distributionsAD[dirTSE *size_Mat];
      }
	  ////////////////////////////////////////////////////////////////////////////////
	  //! - Set neighbor indices (necessary for indirect addressing)
	  //!
	  uint ke   = k;
      uint kw   = neighborX[k];
      uint kn   = k;
      uint ks   = neighborY[k];
      uint kt   = k;
      uint kb   = neighborZ[k];
      uint ksw  = neighborY[kw];
      uint kne  = k;
      uint kse  = ks;
      uint knw  = kw;
      uint kbw  = neighborZ[kw];
      uint kte  = k;
      uint kbe  = kb;
      uint ktw  = kw;
      uint kbs  = neighborZ[ks];
      uint ktn  = k;
      uint kbn  = kb;
      uint kts  = ks;
      uint ktse = ks;
      uint kbnw = kbw;
      uint ktnw = kw;
      uint kbse = kbs;
      uint ktsw = ksw;
      uint kbne = kb;
      uint ktne = k;
      uint kbsw = neighborZ[ksw];
	  ////////////////////////////////////////////////////////////////////////////////
	  //! - Set local distributions
	  //!
	  real mfcbb = (distAD.f[dirE   ])[ke  ];
	  real mfabb = (distAD.f[dirW   ])[kw  ];
	  real mfbcb = (distAD.f[dirN   ])[kn  ];
	  real mfbab = (distAD.f[dirS   ])[ks  ];
	  real mfbbc = (distAD.f[dirT   ])[kt  ];
	  real mfbba = (distAD.f[dirB   ])[kb  ];
	  real mfccb = (distAD.f[dirNE  ])[kne ];
	  real mfaab = (distAD.f[dirSW  ])[ksw ];
	  real mfcab = (distAD.f[dirSE  ])[kse ];
	  real mfacb = (distAD.f[dirNW  ])[knw ];
	  real mfcbc = (distAD.f[dirTE  ])[kte ];
	  real mfaba = (distAD.f[dirBW  ])[kbw ];
	  real mfcba = (distAD.f[dirBE  ])[kbe ];
	  real mfabc = (distAD.f[dirTW  ])[ktw ];
	  real mfbcc = (distAD.f[dirTN  ])[ktn ];
	  real mfbaa = (distAD.f[dirBS  ])[kbs ];
	  real mfbca = (distAD.f[dirBN  ])[kbn ];
	  real mfbac = (distAD.f[dirTS  ])[kts ];
	  real mfbbb = (distAD.f[dirREST])[k   ];
	  real mfccc = (distAD.f[dirTNE ])[ktne];
	  real mfaac = (distAD.f[dirTSW ])[ktsw];
	  real mfcac = (distAD.f[dirTSE ])[ktse];
	  real mfacc = (distAD.f[dirTNW ])[ktnw];
	  real mfcca = (distAD.f[dirBNE ])[kbne];
	  real mfaaa = (distAD.f[dirBSW ])[kbsw];
	  real mfcaa = (distAD.f[dirBSE ])[kbse];
	  real mfaca = (distAD.f[dirBNW ])[kbnw];
      //////////////////////////////////////////////////////////////////////////
	  //! - Calculate concentration using pyramid summation for low round-off errors as in Eq. (J1)-(J3) \ref
	  //! <a href="https://doi.org/10.1016/j.camwa.2015.05.001"><b>[ M. Geier et al. (2015), DOI:10.1016/j.camwa.2015.05.001 ]</b></a>
	  //!
	  concentration[k] =
	   ((((mfccc + mfaaa) + (mfaca + mfcac)) + ((mfacc + mfcaa)   + (mfaac + mfcca))) +
      	(((mfbac + mfbca) + (mfbaa + mfbcc)) + ((mfabc + mfcba)   + (mfaba + mfcbc)) + ((mfacb + mfcab) + (mfaab + mfccb))) +
      	 ((mfabb + mfcbb) + (mfbab + mfbcb)  +  (mfbba + mfbbc))) +  mfbbb;

   }
}





